@@ -17,11 +17,14 @@
 
    You should have received a copy of the GNU Lesser General Public
    License along with the m17n library; if not, write to the Free
-   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    02111-1307, USA.  */
 
 #include <config.h>
 #include <stdlib.h>
+#include <string.h>
+#include <ctype.h>
+#include <locale.h>
 #include "m17n.h"
 #include "m17n-misc.h"
 #include "internal.h"
@@ -34,6 +37,7 @@
 
 static MPlist *language_list;
 static MPlist *script_list;
+static MPlist *langname_list;
 
 static MPlist *
 load_lang_script_list (MSymbol tag0, MSymbol tag1, MSymbol tag2, MSymbol tag3)
@@ -45,18 +49,18 @@ load_lang_script_list (MSymbol tag0, MSymbol tag1, MSymbol tag2, MSymbol tag3)
       || ! (plist = mdatabase_load (mdb)))
     return NULL;
   /* Check at least if the plist is ((SYMBOL ...) ...).  */
-  MPLIST_DO (pl, plist)
+  for (pl = plist; ! MPLIST_TAIL_P (pl);)
     {
       if (! MPLIST_PLIST_P (pl))
-	break;
-      p = MPLIST_PLIST (pl);
-      if (! MPLIST_SYMBOL_P (p))
-	break;
-    }
-  if (! MPLIST_TAIL_P (pl))
-    {
-      M17N_OBJECT_UNREF (plist);
-      return NULL;
+	mplist__pop_unref (pl);
+      else
+	{
+	  p = MPLIST_PLIST (pl);
+	  if (! MPLIST_SYMBOL_P (p))
+	    mplist__pop_unref (pl);
+	  else
+	    pl = MPLIST_NEXT (pl);
+	}
     }
   return plist;
 }
@@ -88,20 +92,71 @@ init_script_list (void)
   return 0;
 }
 
+static MPlist *
+load_lang_name (MSymbol target3, MSymbol target2)
+{
+  MPlist *plist, *pl, *p;
+
+  plist = mplist ();
+  mplist_add (plist, Msymbol, target3);
+  pl = mdatabase_list (Mlanguage, Mname, target3, Mnil);
+  if (! pl && target2 != Mnil)
+    pl = mdatabase_list (Mlanguage, Mname, target2, Mnil);
+  if (pl)
+    {
+      MPLIST_DO (p, pl)
+	{
+	  MDatabase *mdb = MPLIST_VAL (p);
+	  MPlist *p0 = mdatabase_load (mdb), *p1, *territories;
+	  MSymbol script = mdatabase_tag (mdb)[3];
+	  
+	  if (MPLIST_PLIST_P (p0))
+	    {
+	      p1 = MPLIST_PLIST (p0);
+	      if (MPLIST_SYMBOL_P (p1) && MPLIST_SYMBOL (p1) == Mlanguage)
+		{
+		  p1 = MPLIST_NEXT (MPLIST_NEXT (MPLIST_NEXT (MPLIST_NEXT (p1))));
+		  territories = p1;
+		  while (! MPLIST_TAIL_P (p1))
+		    {
+		      if (MPLIST_SYMBOL_P (p1))
+			p1 = MPLIST_NEXT (p1);
+		      else
+			mplist__pop_unref (p1);
+		    }
+		  M17N_OBJECT_REF (territories);
+		  mplist__pop_unref (p0);
+		}
+	      else
+		territories = mplist ();
+	      mplist_push (p0, Mplist, territories);
+	      M17N_OBJECT_UNREF (territories);
+	      mplist_push (p0, Msymbol, script);
+	      mplist_add (plist, Mplist, p0);
+	      M17N_OBJECT_UNREF (p0);
+	    }
+	}
+      M17N_OBJECT_UNREF (pl);
+    }
+  mplist_push (langname_list, Mplist, plist);
+  M17N_OBJECT_UNREF (plist);
+  return plist;
+}
+
 
 /* Internal API */
 
 int
 mlang__init ()
 {
-  msymbol_put (Mlanguage, Mtext_prop_serializer,
-	       (void *) msymbol__serializer);
-  msymbol_put (Mlanguage, Mtext_prop_deserializer,
-	       (void *) msymbol__deserializer);
+  msymbol_put_func (Mlanguage, Mtext_prop_serializer,
+		    M17N_FUNC (msymbol__serializer));
+  msymbol_put_func (Mlanguage, Mtext_prop_deserializer,
+		    M17N_FUNC (msymbol__deserializer));
   Miso639_2 = msymbol ("iso639-2");
   Miso639_1 = msymbol ("iso639-1");
 
-  language_list = script_list = NULL;
+  language_list = script_list = langname_list = NULL;
   return 0;
 }
 
@@ -109,9 +164,8 @@ void
 mlang__fini (void)
 {
   M17N_OBJECT_UNREF (language_list);
-  language_list = NULL;
   M17N_OBJECT_UNREF (script_list);
-  script_list = NULL;
+  M17N_OBJECT_UNREF (langname_list);
 }
 
 /*=*/
@@ -163,19 +217,25 @@ mlanguage__info (MSymbol language)
     {
       MPlist *pl = MPLIST_PLIST (plist);
 
+      if (MPLIST_SYMBOL (pl) == language)
+	return pl;
+      if (MPLIST_TAIL_P (pl))
+	continue;
+      pl = MPLIST_NEXT (pl);
       if (MPLIST_SYMBOL_P (pl) && MPLIST_SYMBOL (pl) == language)
 	return MPLIST_PLIST (plist);
-      if (! MPLIST_TAIL_P (pl))
+      if (MPLIST_TAIL_P (pl))
+	continue;
+      pl = MPLIST_NEXT (pl);
+      if (MPLIST_MTEXT_P (pl))
 	{
-	  pl = MPLIST_NEXT (pl);
-	  if (MPLIST_SYMBOL_P (pl) && MPLIST_SYMBOL (pl) == language)
+	  MText *mt = MPLIST_MTEXT (pl);
+
+	  if (mtext_nbytes (mt) == MSYMBOL_NAMELEN (language)
+	      && strncasecmp ((char *) MTEXT_DATA (MPLIST_MTEXT (pl)),
+			      MSYMBOL_NAME (language),
+			      MSYMBOL_NAMELEN (language)) == 0)
 	    return MPLIST_PLIST (plist);
-	  if (! MPLIST_TAIL_P (pl))
-	    {
-	      pl = MPLIST_NEXT (pl);
-	      if (MPLIST_SYMBOL_P (pl) && MPLIST_SYMBOL (pl) == language)
-		return MPLIST_PLIST (plist);
-	    }
 	}
     }
   return NULL;
@@ -291,15 +351,15 @@ MSymbol Miso639_1, Miso639_2;
     mscript_list ().  */
 
 /***ja
-    @brief 3$BJ8;z8@8l%3!<%I$r%j%9%H$9$k(B.
+    @brief 3ʸ쥳ɤꥹȤ.
 
-    $B4X?t(B mlanguage_list () $B$O!"@07A<0(B (well-formed) plist $B$rJV$9!#3F%-!<(B
-    $B$O(B #Msymbol $B$G$"$j!"8D!9$NCM$O(B ISO639-2 $B$KDj$a$i$l$?(B3$BJ8;z8@8l%3!<(B
-    $B%I$rL>A0$H$9$k%7%s%\%k$G$"$k!#(B
+    ؿ mlanguage_list () ϡ (well-formed) plist ֤ƥ
+     #Msymbol Ǥꡢġͤ ISO639-2 줿3ʸ쥳
+    ɤ̾Ȥ륷ܥǤ롣
 
     @return
-    $B$3$N4X?t$,JV$9(B plist $B$O!"8F$S=P$7B&$,(B m17n_object_unref () $B$r;H$C(B
-    $B$F2rJ|$9$kI,MW$,$"$k!#(B
+    δؿ֤ plist ϡƤӽФ¦ m17n_object_unref () Ȥ
+    Ʋɬפ롣
 
     @seealso
     mscript_list ().  */
@@ -342,25 +402,25 @@ mlanguage_list (void)
     symbol.  Otherwise, it returns #Mnil.
 
     @seealso
-    mlanguage_name (), mlanguage_text ().  */
+    mlanguage_name_list (), mlanguage_text ().  */
 
 /***ja
-    @brief $B8@8l%3!<%I$rF@$k(B.
+    @brief 쥳ɤ.
 
-    $B4X?t(B mlanguage_code () $B$O!"(B$LANGUAGE $B$KBP1~$7$?(B ISO-639 $B8@8l%3!<%I(B
-    $B$,L>A0$G$"$k$h$&$J%7%s%\%k$rJV$9!#(B$LANGUAGE $B$O%7%s%\%k$G$"$j!"$=$N(B
-    $BL>A0$O!"(BISO639-2 3$BJ8;z8@8l%3!<%I!"(BISO639-1 2$BJ8;z8@8l%3!<%I!"1Q8lL>!"(B
-    $B$N$$$:$l$+$G$"$k!#(B
+    ؿ mlanguage_code () ϡ$LANGUAGE б ISO-639 쥳
+    ̾Ǥ褦ʥܥ֤$LANGUAGE ϥܥǤꡢ
+    ̾ϡISO639-2 3ʸ쥳ɡISO639-1 2ʸ쥳ɡѸ̾
+    Τ줫Ǥ롣
 
-    $LEN $B$OJV$5$l$k8@8l%3!<%I$N<oN`$r7hDj$9$k!#(B$LEN $B$,(B3$B$N>l9g$O(B 
-    ISO639-2 3$BJ8;z8@8l%3!<%I$,JV$5$l$k!#(B2$B$N>l9g$O!"$b$7Dj5A$5$l$F$$$l(B
-    $B$P(B ISO639-1 2$BJ8;z8@8l%3!<%I$,!"$=$&$G$J$1$l$P(B #Mnil $B$,JV$5$l$k!#(B0 
-    $B$N>l9g$O!"$b$7Dj5A$5$l$F$$$l$P(B2$BJ8;z%3!<%I$,!"$=$&$G$J$1$l$P(B3$BJ8;z%3!<(B
-    $B%I$,JV$5$l$k!#(B
+    $LEN ֤쥳ɤμꤹ롣$LEN 3ξ 
+    ISO639-2 3ʸ쥳ɤ֤롣2ξϡ⤷Ƥ
+     ISO639-1 2ʸ쥳ɤǤʤ #Mnil ֤롣0 
+    ξϡ⤷Ƥ2ʸɤǤʤ3ʸ
+    ɤ֤롣
 
     @return
-    $B$b$7>pJs$,F@$i$l$l$P!"$3$N4X?t$O(B #Mnil $B0J30$N%7%s%\%k$rJV$9!#$=$&(B
-    $B$G$J$1$l$P(B #Mnil $B$rJV$9!#(B
+    ⤷Сδؿ #Mnil ʳΥܥ֤
+    Ǥʤ #Mnil ֤
 
     @seealso
     mlanguage_name (), mlanguage_text ().  */
@@ -387,48 +447,156 @@ mlanguage_code (MSymbol language, int len)
 /*=*/
 
 /***en
-    @brief Get an English language name.
+    @brief Return the language names written in the specified language.
 
-    The mlanguage_name () function returns a symbol whose name is an
-    English name of $LANGUAGE.  $LANGUAGE is a symbol whose name is an
-    ISO639-2 3-letter language code, an ISO639-1 2-letter language
-    codes, or an English word.
+    The mlanguage_name_list () function returns a plist of LANGUAGE's
+    names written in TARGET language.  SCRIPT and TERRITORY, if not #Mnil,
+    specifies which script and territory to concern at first.
+
+    LANGUAGE and TARGET must be a symbol whose name is an ISO639-2
+    3-letter language code or an ISO639-1 2-letter language codes.
+    TARGET may be #Mnil, in which case, the language of the current
+    locale is used.  If locale is not set or is C, English is used.
+
+    SCRIPT and TERRITORY must be a symbol whose name is a script and
+    territory name of a locale (e.g. "TW", "SG") respectively.
 
     @return
-    If the information is available, this function returns a non-#Mnil
-    symbol.  Otherwise, it returns #Mnil.
+    If the translation is available, this function returns a non-empty
+    plist.  The first element has key #MText and the value is an
+    M-text of a translated language name.  If the succeeding elements
+    also have key #MText, their values are M-texts of alternate
+    translations.
+
+    If no translation is available, @c NULL is returned.
+
+    The returned plist should not be modified nor freed.
 
     @seealso
     mlanguage_code (), mlanguage_text ().  */
 
-/***ja
-    @brief $B8@8l$N1Q8lL>$rF@$k(B.
-
-    $B4X?t(B mlanguage_name () $B$O!"(B$LANGUAGE $B$N1Q8lL>$rL>A0$H$9$k$h$&$J%7(B
-    $B%s%\%k$rJV$9!#(B$LANGUAGE $B$O%7%s%\%k$G$"$j!"$=$NL>A0$O!"(BISO639-2 3$BJ8(B
-    $B;z8@8l%3!<%I!"(BISO639-1 2$BJ8;z8@8l%3!<%I!"1Q8lL>!"$N$$$:$l$+$G$"$k!#(B
-
-    @return
-    $B5a$a$F$$$k>pJs$,F@$i$l$k$J$i!"$3$N4X?t$O(B #Mnil $B0J30$N%7%s%\%k$rJV(B
-    $B$9!#$=$&$G$J$1$l$P(B #Mnil $B$rJV$9!#(B
-
-    @seealso
-    mlanguage_code (), mlanguage_text ().  */
-
-MSymbol
-mlanguage_name (MSymbol language)
+MPlist *
+mlanguage_name_list (MSymbol language, MSymbol target,
+		     MSymbol script, MSymbol territory)
 {
-  MPlist *plist = mlanguage__info (language);
+  MPlist *plist, *pl;
+  MSymbol language2, target2;
 
-  if (! plist)			/* 3-letter code */
-    return Mnil;
-  plist = MPLIST_NEXT (plist);	/* 2-letter code */
+  plist = mlanguage__info (language);
+  if (! plist)
+    return NULL;
+  language = MPLIST_SYMBOL (plist);
+  language2 = MPLIST_SYMBOL (MPLIST_NEXT (plist));
+  if (target != Mnil)
+    {
+      plist = mlanguage__info (target);
+      if (! plist)
+	return NULL;
+      target = MPLIST_SYMBOL (plist);
+      target2 = MPLIST_SYMBOL (MPLIST_NEXT (plist));
+    }
+  else
+    {
+      MLocale *locale = mlocale_set (LC_MESSAGES, NULL);
+
+      if (! locale)
+	{
+	  target = msymbol ("eng");
+	  target2 = msymbol ("en");
+	  script = territory = Mnil;
+	}
+      else
+	{
+	  target = mlocale_get_prop (locale, Mlanguage);
+	  plist = mlanguage__info (target);
+	  if (! plist)
+	    return NULL;
+	  target = MPLIST_SYMBOL (plist);
+	  target2 = MPLIST_SYMBOL (MPLIST_NEXT (plist));
+	  script = mlocale_get_prop (locale, Mscript);
+	  territory = mlocale_get_prop (locale, Mterritory);
+	}
+    }
+
+  /* Now both LANGUAGE and TARGET are 3-letter codes.  */
+  if (! langname_list)
+    langname_list = mplist ();
+  plist = mplist__assq (langname_list, target);
+  if (plist)
+    plist = MPLIST_PLIST (plist);
+  else
+    plist = load_lang_name (target, target2);
+
+  /* PLIST = (TARGET (SCRIPT (TERRITORY ...) (LANG-CODE NAME ...) ...) ...) */
+  plist = MPLIST_NEXT (plist);
   if (MPLIST_TAIL_P (plist))
-    return Mnil;
-  plist = MPLIST_NEXT (plist);	/* english name */
-  if (! MPLIST_SYMBOL_P (plist))
-    return Mnil;
-  return MPLIST_SYMBOL (plist);
+    return NULL;
+
+  MPLIST_DO (pl, plist)
+    {
+      MPlist *p = MPLIST_PLIST (pl), *p0;
+
+      if (MPLIST_SYMBOL (p) == script
+	  && (territory == Mnil
+	      || mplist_find_by_value (MPLIST_PLIST (MPLIST_NEXT (p)),
+				       territory))
+	  && (p = MPLIST_NEXT (MPLIST_NEXT (p)))
+	  && ((p0 = mplist__assq (p, language))
+	      || (p0 = mplist__assq (p, language2))))
+	{
+	  plist = p0;
+	  break;
+	}
+    }
+  if (MPLIST_TAIL_P (pl))
+    {
+      MPLIST_DO (pl, plist)
+	{
+	  MPlist *p = MPLIST_NEXT (MPLIST_PLIST (pl)), *p0;
+
+	  if ((territory == Mnil
+	       || mplist_find_by_value (MPLIST_VAL (p), territory))
+	      && (p = MPLIST_NEXT (MPLIST_NEXT (p)))
+	      && ((p0 = mplist__assq (p, language))
+		  || (p0 = mplist__assq (p, language2))))
+	    {
+	      plist = p0;
+	      break;
+	    }
+	}
+      if (MPLIST_TAIL_P (pl))
+	{
+	  MPLIST_DO (pl, plist)
+	    {
+	      MPlist *p = MPLIST_NEXT (MPLIST_NEXT (MPLIST_PLIST (pl))), *p0;
+
+	      if ((p0 = mplist__assq (p, language))
+		  || (p0 = mplist__assq (p, language2)))
+		{
+		  plist = p0;
+		  break;
+		}
+	    }
+	  if (MPLIST_TAIL_P (pl))
+	    return NULL;
+	}
+    }
+
+  plist = MPLIST_NEXT (MPLIST_PLIST (plist));
+  if (territory != Mnil)
+    {
+      MPLIST_DO (pl, MPLIST_NEXT (plist))
+	if (MPLIST_SYMBOL_P (pl) && MPLIST_SYMBOL (pl) == territory)
+	  break;
+      if (! MPLIST_TAIL_P (pl)
+	  && (pl = MPLIST_NEXT (pl))
+	  && MPLIST_MTEXT_P (pl))
+	return pl;
+    }
+  MPLIST_DO (plist, plist)
+    if (MPLIST_MTEXT_P (plist))
+      break;
+  return (MPLIST_MTEXT_P (plist) ? plist : NULL);
 }
 
 /*=*/
@@ -452,16 +620,16 @@ mlanguage_name (MSymbol language)
     mlanguage_code (), mlanguage_name ().  */
 
 /***ja
-    @brief $BM?$($i$l$?8@8l<+?H$G=q$+$l$?8@8lL>$rJV$9(B.
+    @brief Ϳ줿켫Ȥǽ񤫤줿֤̾.
 
-    $B4X?t(B mlanguage_text () $B$O!"8@8l(B $LANGUAGE $B$G=q$+$l$?(B $LANGUAGE $B$N(B
-    $BL>A0$r(B M-text $B$N7A<0$GJV$9!#$=$N8@8l$NBeI=E*$JJ8;z$,$o$+$C$F$$$k>l(B
-    $B9g$O!"JV$5$l$k(B M-text $B$N3FJ8;z$K!"%-!<$,(B #Mtext $B$GCM$,$=$NBeI=E*$J(B
-    $BJ8;z$r4^$`(B M-text $B$G$"$k$h$&$J%F%-%9%H%W%m%Q%F%#$,IU2C$5$l$k!#(B
+    ؿ mlanguage_text () ϡ $LANGUAGE ǽ񤫤줿 $LANGUAGE 
+    ̾ M-text η֤θɽŪʸ狼äƤ
+    ϡ֤ M-text γʸˡ #Mtext ͤɽŪ
+    ʸޤ M-text Ǥ褦ʥƥȥץѥƥղä롣
 
     @return
-    $B5a$a$k>pJs$,F@$i$l$?>l9g!"$3$N4X?t$,JV$9(B M-text $B$rJQ99$7$?$j2rJ|$7(B
-    $B$?$j$7$F$O$$$1$J$$!#>pJs$,F@$i$l$J$+$C$?>l9g$O(B @c NULL $B$,JV$5$l$k!#(B
+    줿硢δؿ֤ M-text ѹ
+    ꤷƤϤʤʤä @c NULL ֤롣
 
     @seealso
     mlanguage_code (), mlanguage_name ().  */
@@ -508,15 +676,15 @@ mlanguage_text (MSymbol language)
     mscript_language_list (), mlanguage_list ().  */
 
 /***ja
-    @brief $B%9%/%j%W%HL>$r%j%9%H$9$k(B.
+    @brief ץ̾ꥹȤ.
 
-    $B4X?t(B mscript_list () $B$O!"@07A<0(B (well-formed) plist $B$rJV$9!#3F%-!<(B
-    $B$O(B #Msymbol $B$G$"$j!"8D!9$NCM$O%9%/%j%W%HL>$rL>A0$H$9$k%7%s%\%k$G$"(B
-    $B$k!#(B
+    ؿ mscript_list () ϡ (well-formed) plist ֤ƥ
+     #Msymbol Ǥꡢġͤϥץ̾̾Ȥ륷ܥǤ
+    롣
 
     @return
-    $B$3$N4X?t$,JV$9(B plist $B$O!"8F$S=P$7B&$,(B m17n_object_unref () $B$r;H$C(B
-    $B$F2rJ|$9$kI,MW$,$"$k!#(B
+    δؿ֤ plist ϡƤӽФ¦ m17n_object_unref () Ȥ
+    Ʋɬפ롣
 
     @seealso
     mscript_language_list (), mlanguage_list ().  */
@@ -560,17 +728,17 @@ mscript_list (void)
     mscript_list (), mlanguage_list ().  */
 
 /***ja
-    @brief $BM?$($i$l$?%9%/%j%W%H$rMQ$$$k8@8l$r%j%9%H$9$k(B.
+    @brief Ϳ줿ץȤѤꥹȤ.
 
-    $B4X?t(B mscript_language_list () $B$O!"(B$SCRIPT $B$rMQ$$$k8@8l$r%j%9%H$9$k!#(B
-    $SCRIPT $B$O%7%s%\%k$G!"$=$NL>A0$O(B Unicode Character Database $B$K<($5(B
-    $B$l$F$$$k%9%/%j%W%HL>$r$9$Y$F>.J8;z$K$7$?$b$N$G$"$k!#(B
+    ؿ mscript_language_list () ϡ$SCRIPT ѤꥹȤ롣
+    $SCRIPT ϥܥǡ̾ Unicode Character Database ˼
+    Ƥ륹ץ̾򤹤٤ƾʸˤΤǤ롣
 
-    @return $B$3$N4X?t$O!"@07A<0(B (well-formed) plist $B$rJV$9!#3F%-!<$O(B 
-    #Msymbol $B$G$"$j!"8D!9$NCM$O(B ISO639-1 $B$KDj$a$i$l$?(B2$BJ8;z8@8l%3!<%I(B
-    ($BDj5A$5$l$F$$$J$$>l9g$O(B ISO639-2 $B$KDj$a$i$l$?(B3$BJ8;z8@8l%3!<%I(B) $B$rL>(B
-    $BA0$H$9$k%7%s%\%k$G$"$k!#JV$5$l$k(B plist $B$OJQ99$7$?$j2rJ|$7$?$j$7$F(B
-    $B$O$J$i$J$$!#(B$SCRIPT $B$,L$CN$N>l9g$O(B @c NULL $B$,JV$5$l$k!#(B
+    @return δؿϡ (well-formed) plist ֤ƥ 
+    #Msymbol Ǥꡢġͤ ISO639-1 줿2ʸ쥳
+    (Ƥʤ ISO639-2 줿3ʸ쥳) ̾
+    Ȥ륷ܥǤ롣֤ plist ѹꤷ
+    Ϥʤʤ$SCRIPT ̤Τξ @c NULL ֤롣
 
     @seealso
     mscript_list (), mlanguage_list ().  */
@@ -587,3 +755,74 @@ mscript_language_list (MSymbol script)
     return MPLIST_PLIST (plist);
   return NULL;
 }
+
+/*** @} */
+/*=*/
+/***en
+    @name Obsolete functions
+*/
+/***ja
+    @name Obsolete ʴؿ
+*/
+/*** @{ */
+
+/***en
+    @brief Get an English language name.
+
+    This function is obsolete.  Use mlanguage_name_list () instead.
+
+    The mlanguage_name () function returns a symbol whose name is an
+    English name of $LANGUAGE.  $LANGUAGE is a symbol whose name is an
+    ISO639-2 3-letter language code, an ISO639-1 2-letter language
+    codes, or an English word.
+
+    @return
+    If the information is available, this function returns a non-#Mnil
+    symbol.  Otherwise, it returns #Mnil.
+
+    @seealso
+    mlanguage_code (), mlanguage_text ().  */
+
+/***ja
+    @brief αѸ̾.
+
+    ؿ mlanguage_name () ϡ$LANGUAGE αѸ̾̾Ȥ褦ʥ
+    ܥ֤$LANGUAGE ϥܥǤꡢ̾ϡISO639-2 3ʸ
+    쥳ɡISO639-1 2ʸ쥳ɡѸ̾Τ줫Ǥ롣
+
+    @return
+    Ƥʤ顢δؿ #Mnil ʳΥܥ
+    Ǥʤ #Mnil ֤
+
+    @seealso
+    mlanguage_code (), mlanguage_text ().  */
+
+MSymbol
+mlanguage_name (MSymbol language)
+{
+  MPlist *plist = mlanguage__info (language);
+  MText *mt;
+  char *str;
+
+  if (! plist)			/* 3-letter code */
+    return Mnil;
+  plist = MPLIST_NEXT (plist);	/* 2-letter code */
+  if (MPLIST_TAIL_P (plist))
+    return Mnil;
+  plist = MPLIST_NEXT (plist);	/* english name */
+  if (! MPLIST_MTEXT_P (plist))
+    return Mnil;
+  mt = MPLIST_MTEXT (plist);
+  str = alloca (mtext_nbytes (mt));
+  memcpy (str, MTEXT_DATA (mt), mtext_nbytes (mt));
+  str[0] = tolower (str[0]);
+  return msymbol__with_len (str, mtext_nbytes (mt));
+}
+
+/*=*/
+
+/*
+  Local Variables:
+  coding: euc-japan
+  End:
+*/
