@@ -1,5 +1,5 @@
 /* m17n-core.c -- body of the CORE API.
-   Copyright (C) 2003, 2004
+   Copyright (C) 2003, 2004, 2005, 2006, 2007
      National Institute of Advanced Industrial Science and Technology (AIST)
      Registration Number H15PRO112
 
@@ -17,7 +17,7 @@
 
    You should have received a copy of the GNU Lesser General Public
    License along with the m17n library; if not, write to the Free
-   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    02111-1307, USA.  */
 
 /***en
@@ -31,88 +31,98 @@
     <ol>
     <li> CORE API
 
-    It provides basic modules to handle M-texts.  They do not require
-    the m17n database.  To use this API, an application program must
-    include <m17n-core<EM></EM>.h> and be linked by -lm17n-core.
+    It provides basic modules to handle M-texts.  To use this API, an
+    application program must include <m17n-core<EM></EM>.h> and be
+    linked with -lm17n-core.
 
     <li> SHELL API
 
-    It provides modules that utilize the m17n database (code
-    conversion, character property, etc).  They load various kinds of
+    It provides modules for character properties, character set
+    handling, code conversion, etc.  They load various kinds of
     data from the database on demand.  To use this API, an application
-    program must include <m17n<EM></EM>.h> and be linked by
+    program must include <m17n<EM></EM>.h> and be linked with
     -lm17n-core -lm17n.
 
     When you use this API, CORE API is also available.
 
+    <li> FLT API
+
+    It provides modules for text shaping using @ref mdbFLT.  To use
+    this API, an application program must include <m17n<EM></EM>.h>
+    and be linked with -lm17n-core -lm17n-flt.
+
+    When you use this API, CORE API is also available.
+
     <li> GUI API
 
     It provides GUI modules such as drawing and inputting M-texts on a
-    graphic device.  The API itself is independent on a graphic device
-    but most functions require an argument MFrame which is created for
-    a specific type of graphic device.  Currently, a null device, X
-    Window System, and an image date (gdImagePtr) of GD library are
-    supported as a graphic device.
+    graphic device.  This API itself is independent of graphic
+    devices, but most functions require an argument MFrame that is
+    created for a specific type of graphic devices.  The currently
+    supported graphic devices are null device, the X Window System,
+    and image data (gdImagePtr) of the GD library.
 
-    On a frame of the null device, you can't draw text nor use an
-    input method (but, for instance, the function mdraw_glyph_list ()
-    is available).
+    On a frame of a null device, you cannot draw text nor use input
+    methods.  However, functions like mdraw_glyph_list (), etc. are
+    available.
 
     On a frame of the X Window System, you can use the whole GUI API.
 
-    On a frame of GD library, you can use all drawing API but can't
-    use input method.
+    On a frame of the GD library, you can use all drawing API but
+    cannot use input methods.
 
-    To use this device, an application program must include
-    <m17n-gui<EM></EM>.h> and be linked by -lm17n-core -lm17n
+    To use this API, an application program must include
+    <m17n-gui<EM></EM>.h> and be linked with -lm17n-core -lm17n
     -lm17n-gui.
 
-    When you use this API, CORE and SHELL API are also available.
+    When you use this API, CORE, SHELL, and FLT APIs are also
+    available.
 
     <li> MISC API
 
     It provides miscellaneous functions to support error handling and
-    debugging.  This API cannot be used by itself, but with one or
-    more APIs listed above.  To use the API, an application program
-    must include <m17n-misc<EM></EM>.h> in addition to one of a header
-    file described above.
+    debugging.  This API cannot be used standalone; it must be used
+    with one or more APIs listed above.  To use this API, an
+    application program must include <m17n-misc<EM></EM>.h> in
+    addition to one of the header files described above.
 
     </ol>
 
     See also the section @ref m17n-config "m17n-config(1)".
 
-    <em>ENVIRONMENT VARIABLE</em>
+    <em>ENVIRONMENT VARIABLES</em>
 
-    The m17n library pays attention to these environment variables.
+    The m17n library pays attention to the following environment
+    variables.
 
     <ul>
     <li> @c M17NDIR
 
-    Name of a directory that contains data of the m17n database.  See
-    @ref m17nDatabase for more details.
+    The name of the directory that contains data of the m17n database.
+    See @ref m17nDatabase for details.
 
-    <li> @c MDEBUG_XXXX
+    <li> @c MDEBUG_XXX
 
-    Environment variables whose name start by "MDEBUG_" controls
-    printing of debug information.  See @ref m17nDebug for more
-    details.
+    Environment variables whose names start with "MDEBUG_" control
+    debug information output.  See @ref m17nDebug for details.
 
     </ul>
 
     <em>API NAMING CONVENTION</em>
 
-    The library exports functions, variables, macros, and types.  All
-    of them start by the letter 'm' or 'M' followed by an object name
-    (e.g. "symbol" and "plist", but "mtext" object is given the name
-    "text" to avoid double 'm' at the head) or a module name
-    (e.g. draw, input).
-    
+    The m17n library exports functions, variables, macros, and types.
+    All of them start with the letter 'm' or 'M', and are followed by
+    an object name (e.g. "symbol", "plist") or a module name
+    (e.g. draw, input).  Note that the name of M-text objects start
+    with "mtext" and not with "mmtext".
+
     <ul>
 
     <li> functions -- mobject () or mobject_xxx ()
 
-    They start with 'm' followed by lower case object name.  For
-    example, msymbol (), mtext_ref_char (), mdraw_text ().
+    They start with 'm' and are followed by an object name in lower
+    case.  Words are separated by '_'.  For example, msymbol (),
+    mtext_ref_char (), mdraw_text ().
 
     <li> non-symbol variables -- mobject, or mobject_xxx
     
@@ -120,18 +130,20 @@
 
     <li> symbol variables -- Mname
 
-    Variables of type MSymbol start with 'M' followed by their names
-    (e.g. Mlanguage (name is "language"), Miso_2022 (name is
-    "iso-2022").
+    Variables of the type MSymbol start with 'M' and are followed by
+    their names.  Words are separated by '_'.  For example, Mlanguage
+    (the name is "language"), Miso_2022 (the name is "iso-2022").
 
     <li> macros -- MOBJECT_XXX
 
-    They start by 'M' followed by upper case object names.
+    They start with 'M' and are followed by an object name in upper
+    case.  Words are separated by '_'.
 
     <li> types -- MObject or MObjectXxx
 
-    They start by 'M' followed by capitalized object names (e.g.
-    MConverter, MInputDriver).
+    They start with 'M' and are followed by capitalized object names.
+    Words are concatenated directly and no '_' are used.  For example,
+    MConverter, MInputDriver.
 
     </ul>
 
@@ -148,44 +160,62 @@
     <ol>
     <li>  API
 
-    M-text 򰷤δŪʥ⥸塼󶡤롣ʬ API 
-    m17n ǡ١ɬפȤʤѤ뤿ˤϡץꥱץ
-    <m17n-core<EM></EM>.h>  include  -lm17n-core ǥ󥯤ʤƤϤʤʤ
+    M-text 򰷤δŪʥ⥸塼󶡤롣
+     API Ѥ뤿ˤϡץꥱץ
+    <m17n-core<EM></EM>.h>  include  -lm17n-core
+    ǥ󥯤ʤƤϤʤʤ
 
     <li>  API
 
-    m17n ǡ١Ѥ⥸塼ʥѴʸץѥƥ
-    󶡤롣⥸塼ϥǡ١ɬפ˱¿ͤʥǡɤ롣
-    Ѥ뤿ˤϡץꥱץ <m17n<EM></EM>.h>  include
-     -lm17n-core -lm17n ǥ󥯤ʤƤϤʤʤ
+    ʸץѥƥʸѴΤΥ⥸塼󶡤롣
+    Υ⥸塼ϡǡ١ɬפ˱¿ͤʥǡɤ롣
+     API Ѥ뤿ˤϡץꥱץ
+    <m17n<EM></EM>.h>  include  -lm17n-core -lm17n
+    ǥ󥯤ʤƤϤʤʤ
 
-     API Ѥݤˤϥ API ѤǤ롣
+     API ѤС API ⼫ưŪ˻ѲǽȤʤ롣
+
+    <li> FLT API
+
+    ʸɽ @ref mdbFLT Ѥ⥸塼󶡤롣 API
+    Ѥ뤿ˤϡץꥱץ <m17n<EM></EM>.h> 
+     include  -lm17n-core -lm17n-flt ǥ󥯤ʤƤϤʤʤ
+
+     API ѤС API ⼫ưŪ˻ѲǽȤʤ롣
 
     <li> GUI API
 
-    M-text 򥰥եåǥХɽϤꤹȤäեå桼󥿥ե⥸塼󶡤롣
-    API ΤϥեåǥХȤΩǤ뤬¿δؿΥեåǥХ˺줿 
-    MFrame ˼롣Ǥϡ̥ǥХX ɥƥࡢ
-    GD 饤֥Υ᡼gdImagePtrˡեåǥХȤƥݡȤƤ롣
+    եåǥХ M-text ɽϤꤹ뤿
+    GUI ⥸塼󶡤롣 API
+    ΤϥեåǥХȤΩǤ뤬
+    ¿δؿΥեåǥХѤ˺줿 
+    MFrame ˼롣
+    ǥݡȤƤ륰եåǥХϡ̥ǥХX
+    ɥƥࡢ GD 饤֥Υ᡼ǡ
+    (gdImagePtr) Ǥ롣
 
-    ̥ǥХǤɽϤǤʤʤȤ
-    mdraw_glyph_list () ʤɤδؿϻѲǽˡ
+    ̥ǥХΥե졼ǤɽϤǤʤ
+    mdraw_glyph_list () ʤɤδؿϻѲǽǤ롣
 
-    X ɥƥǤϤ٤Ƥ GUI API ѤǤ롣
+    X ɥƥΥե졼ǤϤ٤Ƥ GUI API ѤǤ롣
 
-    GD 饤֥Υ᡼ǤϡѤ API Ϥ٤ƻѤǤ뤬ϤϤǤʤ
+    GD 饤֥Υե졼ǤϡѤ API
+    Ϥ٤ƻѤǤ뤬ϤϤǤʤ
 
-    ΥǥХѤ뤿ˤϡץꥱץ
+     API Ѥ뤿ˤϡץꥱץ
     <m17n-gui<EM></EM>.h>  include -lm17n-core -lm17n -lm17n-gui
     ǥ󥯤ʤƤϤʤʤ
 
-     API Ѥݤˤϡ API ȥ API ѤǤ롣
+     API ѤС API API FLT API
+    ⼫ưŪ˻ѲǽȤʤ롣
 
     <li> ¾ API
 
-    顼ǥХåѤΤ¾δؿ󶡤롣 API 
-    ϤǤϻѤǤ嵭¾ΤΤȶ˻ȤѤ뤿ˤϡ嵭Τ줫
-    include ե˲äơ <m17n-misc<EM></EM>.h>  include ʤƤϤʤʤ
+    顼ǥХåѤΤ¾δؿ󶡤롣 API
+    ϤǤϻѤǤ嵭¾ API
+    ȶ˻ȤѤ뤿ˤϡ嵭Τ줫include
+    ե˲äơ <m17n-misc<EM></EM>.h> include
+    ʤƤϤʤʤ
 
     </ol>
 
@@ -193,51 +223,55 @@
 
     @em Ķѿ
 
-    m17n 饤֥ϼδĶѿθ롣
+    m17n 饤֥ϰʲδĶѿ򻲾Ȥ롣
 
     <ul>
     <li> @c M17NDIR
 
-    m17n ǡ١Υǡޤǥ쥯ȥ̾ܺ٤ @ref
+    m17n ǡ١ΥǡǼǥ쥯ȥ̾ܺ٤ @ref
     m17nDatabase ȡ
 
-    <li> @c MDEBUG_XXXX
+    <li> @c MDEBUG_XXX
 
-    "MDEBUG_" ǻϤޤ̾ĴĶѿϥǥХå椹롣ܺ٤ @ref m17nDebug ȡ
+    "MDEBUG_" ǻϤޤ̾ĴĶѿϥǥХåνϤ椹롣
+    ܺ٤ @ref m17nDebug ȡ
 
     </ul>
 
-    @em API @em ̿̾§ ޥ
+    @em API @em ̿̾§
 
-    饤֥ϡؿѿޥ export 롣'm' 
-    ޤ 'M' ΤȤ˥֥̾("symbol"  "plist" ʤɡ 
-    "mtext" ֥ȤϤϤ 'm' νʣ򤱤뤿 "text" 
-    Ѥ롣) ޤϥ⥸塼̾(draw, input ʤ) ³ΤǤ롣
+    m17n 饤֥ϡؿѿޥ export 롣 'm' 
+    ޤ 'M' ΤȤ˥֥̾ ("symbol""plist" ʤ)
+    ޤϥ⥸塼̾ (draw, input ʤ) ³ΤǤ롣
+    M-text ֥Ȥ̾ "mmtext" ǤϤʤ "mtext"
+    ǻϤޤ뤳Ȥա
     
     <ul>
 
     <li> ؿ -- mobject () ޤ mobject_xxx ()
 
-    'm' ΤȤ˾ʸǥ֥̾³ȤСmsymbol (),
+    'm' ΤȤ˾ʸǥ֥̾³ñ֤ '_'
+    Ƕڤ롣ȤСmsymbol (),
      mtext_ref_char (), mdraw_text () ʤɡ
 
     <li> ܥǤʤѿ -- mobject,  ޤ mobject_xxx
     
-    ؿƱ̿̾§˽(Ȥ  mface_large)
+    ؿƱ̿̾§˽Ȥ  mface_large ʤɡ
 
     <li> ܥѿ -- Mname
 
-    MSymbol ѿϡ'M' θ̾³Ȥ Mlanguage (̾ 
-    "language"), Miso_2022 (̾"iso-2022")ʤɡ
+    MSymbol ѿϡ'M' θ̾³ñ֤ '_'
+    Ƕڤ롣Ȥ Mlanguage (̾ "language"), Miso_2022
+    (̾"iso-2022") ʤɡ
 
     <li> ޥ -- MOBJECT_XXX
 
-    'M' θʸǥ֥̾³
+    'M' θʸǥ֥̾³ñ֤ '_' Ƕڤ롣
 
     <li>  -- MObject ޤ MObjectXxx
 
-    'M' θʸǻϤƥ֥̾³ (Ȥ
-    MConverter, MInputDriver)
+    'M' θʸǻϤޤ륪֥̾³ñϢ³ƽ񤫤졢
+    '_' ѤʤȤ MConverter, MInputDriver ʤɡ
 
     </ul>
     
@@ -250,7 +284,7 @@
     of the m17n library.  */
 /***ja
     ޥ #M17NLIB_MAJOR_VERSION  m17n 
-    饤֥Υ᥸㡼СֹͿ롣  */
+    饤֥Υ᥸㡼СֹͿ.  */
 
 #define M17NLIB_MAJOR_VERSION
 
@@ -262,7 +296,7 @@
 
 /***ja
     ޥ #M17NLIB_MINOR_VERSION  m17n 
-    饤֥ΥޥʡСֹͿ롣  */
+    饤֥ΥޥʡСֹͿ.  */
 
 #define M17NLIB_MINOR_VERSION
 
@@ -274,7 +308,7 @@
 
 /***ja
     ޥ #M17NLIB_PATCH_LEVEL  m17n 
-    饤֥Υѥå٥ֹͿ롣  */
+    饤֥Υѥå٥ֹͿ.  */
 
 #define M17NLIB_PATCH_LEVEL
 
@@ -286,7 +320,7 @@
 
 /***ja
     ޥ #M17NLIB_VERSION_NAME  m17n 
-    饤֥ΥС̾ʸȤͿ롣  */
+    饤֥ΥС̾ʸȤͿ.  */
 
 #define M17NLIB_VERSION_NAME
 
@@ -416,11 +450,7 @@ int m17n__core_initialized;
 int m17n__shell_initialized;
 int m17n__gui_initialized;
 
-void *(*mdatabase__finder) (MSymbol tag1, MSymbol tag2,
-			   MSymbol tag3, MSymbol tag4);
-void *(*mdatabase__loader) (void *);
-
-int mdebug__flag;
+int mdebug__flags[MDEBUG_MAX];
 FILE *mdebug__output;
 
 void
@@ -451,19 +481,25 @@ mdebug__print_time ()
   time_stack[time_stack_index - 1] = tv;
 }
 
-#define SET_DEBUG_FLAG(env_name, mask)		\
-  do {						\
-    char *env_value = getenv (env_name);	\
-						\
-    if (env_value)				\
-      {						\
-	if (env_value[0] == '1')		\
-	  mdebug__flag |= (mask);		\
-	else if (env_value[0] == '0')		\
-	  mdebug__flag &= ~(mask);		\
-      }						\
-  } while (0)
+static void
+SET_DEBUG_FLAG (char *env_name, enum MDebugFlag flag)
+{
+  char *env_value = getenv (env_name);
 
+  if (env_value)
+    {
+      int int_value = atoi (env_value);
+
+      if (flag == MDEBUG_ALL)
+	{
+	  int i;
+	  for (i = 0; i < MDEBUG_MAX; i++)
+	    mdebug__flags[i] = int_value;
+	}
+      else
+	mdebug__flags[flag] = int_value;
+    }
+}
 
 void
 mdebug__add_object_array (M17NObjectArray *array, char *name)
@@ -490,11 +526,15 @@ mdebug__unregister_object (M17NObjectArray *array, void *object)
   array->count--;
   if (array->count >= 0)
     {
-      int i = 0;
+      int i;
 
-      while (i < array->used && array->objects[i] != object) i++;
-      if (i < array->used)
-	array->objects[i] = NULL;
+      for (i = array->used - 1; i >= 0 && array->objects[i] != object; i--);
+      if (i >= 0)
+	{
+	  if (i == array->used - 1)
+	    array->used--;
+	  array->objects[i] = NULL;
+	}
       else
 	mdebug_hook ();
     }
@@ -511,7 +551,7 @@ mdebug__unregister_object (M17NObjectArray *array, void *object)
 void
 m17n_init_core (void)
 {
-  int mdebug_mask = MDEBUG_INIT;
+  int mdebug_flag = MDEBUG_INIT;
 
   merror_code = MERROR_NONE;
   if (m17n__core_initialized++)
@@ -519,7 +559,6 @@ m17n_init_core (void)
 
   m17n_memory_full_handler = default_error_handler;
 
-  mdebug__flag = 0;
   SET_DEBUG_FLAG ("MDEBUG_ALL", MDEBUG_ALL);
   SET_DEBUG_FLAG ("MDEBUG_INIT", MDEBUG_INIT);
   SET_DEBUG_FLAG ("MDEBUG_FINI", MDEBUG_FINI);
@@ -527,9 +566,11 @@ m17n_init_core (void)
   SET_DEBUG_FLAG ("MDEBUG_CODING", MDEBUG_CODING);
   SET_DEBUG_FLAG ("MDEBUG_DATABASE", MDEBUG_DATABASE);
   SET_DEBUG_FLAG ("MDEBUG_FONT", MDEBUG_FONT); 
-  SET_DEBUG_FLAG ("MDEBUG_FONT_FLT", MDEBUG_FONT_FLT);
-  SET_DEBUG_FLAG ("MDEBUG_FONT_OTF", MDEBUG_FONT_OTF);
+  SET_DEBUG_FLAG ("MDEBUG_FLT", MDEBUG_FLT);
   SET_DEBUG_FLAG ("MDEBUG_INPUT", MDEBUG_INPUT);
+  /* for backward compatibility... */
+  SET_DEBUG_FLAG ("MDEBUG_FONT_FLT", MDEBUG_FLT);
+  SET_DEBUG_FLAG ("MDEBUG_FONT_OTF", MDEBUG_FLT);
   {
     char *env_value = getenv ("MDEBUG_OUTPUT_FILE");
 
@@ -553,17 +594,18 @@ m17n_init_core (void)
   if  (mplist__init () < 0)
     goto err;
   MDEBUG_PRINT_TIME ("INIT", (stderr, " to initialize plist module."));
+  if (mchar__init () < 0)
+    goto err;
+  MDEBUG_PRINT_TIME ("INIT", (stderr, " to initialize character module."));
   if  (mchartable__init () < 0)
     goto err;
   MDEBUG_PRINT_TIME ("INIT", (stderr, " to initialize chartable module."));
-  if (mtext__init () < 0)
-    goto err;
-  if (mtext__prop_init () < 0)
+  if (mtext__init () < 0 || mtext__prop_init () < 0)
     goto err;
   MDEBUG_PRINT_TIME ("INIT", (stderr, " to initialize mtext module."));
-
-  mdatabase__finder = NULL;
-  mdatabase__loader = NULL;
+  if (mdatabase__init () < 0)
+    goto err;
+  MDEBUG_PRINT_TIME ("INIT", (stderr, " to initialize database module."));
 
 #if ENABLE_NLS
   bindtextdomain ("m17n-lib", GETTEXTDIR);
@@ -583,7 +625,7 @@ m17n_init_core (void)
 void
 m17n_fini_core (void)
 {
-  int mdebug_mask = MDEBUG_FINI;
+  int mdebug_flag = MDEBUG_FINI;
 
   if (m17n__core_initialized == 0
       || --m17n__core_initialized > 0)
@@ -606,7 +648,7 @@ m17n_fini_core (void)
   MDEBUG_POP_TIME ();
   MDEBUG_PRINT_TIME ("FINI", (stderr, " to finalize the core modules."));
   MDEBUG_POP_TIME ();
-  if (mdebug__flag & MDEBUG_FINI)
+  if (mdebug__flags[MDEBUG_FINI])
     report_object_array ();
   msymbol__free_table ();
   if (mdebug__output != stderr)
@@ -1024,11 +1066,9 @@ void (*m17n_memory_full_handler) (enum MErrorCode err);
     <li> MDEBUG_FONT -- If set to 1, print information about fonts
     being selected and opened.
 
-    <li> MDEBUG_FONT_FLT -- If set to 1, print information about which
-    command of Font Layout Table are being executed.
-
-    <li> MDEBUG_FONT_OTF -- If set to 1, print information about which
-    feature of OpenType Layout Table are being executed.
+    <li> MDEBUG_FLT -- If set to 1, 2, or 3, print information about
+    which command of Font Layout Table are being executed.  The bigger
+    number prints the more detailed information.
 
     <li> MDEBUG_INPUT -- If set to 1, print information about how an
     input method is running.
@@ -1073,15 +1113,15 @@ void (*m17n_memory_full_handler) (enum MErrorCode err);
     <li> MDEBUG_DATABASE -- 1 ʤСm17n
     ǡ١ɤ줿ǡˤĤƤξץȤ롣
 
-    <li> MDEBUG_FONT -- 1 ʤС򤵤ƥץ󤵤줿եȤˤĤƤξץȤ롣
+    <li> MDEBUG_FONT -- 1 ʤС򤵤ƥץ󤵤줿եȤˤ
+    ƤξץȤ롣
 
-    <li> MDEBUG_FONT_FLT -- 1 ʤСFont Layout Table 
-    ΤɤΥޥɤ¹椫ˤĤƤΤץȤ롣
+    <li> MDEBUG_FLT -- 12⤷ 3 ʤСFont Layout Table Τ
+    Υޥɤ¹椫ˤĤƤΤץȤ롣礭
+    ץȤ롣
 
-    <li> MDEBUG_FONT_OTF -- 1 ʤСOpenType Layout Table 
-    Τɤ°¹椫ˤĤƤΤץȤ롣
-
-    <li> MDEBUG_INPUT -- 1 ʤС¹ϥ᥽åɤξ֤դƤξץȤ롣
+    <li> MDEBUG_INPUT -- 1 ʤС¹ϥ᥽åɤξ֤դƤ
+    ץȤ롣
 
     <li> MDEBUG_ALL -- 1 ʤС嵭٤Ƥѿ 1 
     ˤΤƱ̤ġ
