@@ -17,20 +17,21 @@
 
    You should have received a copy of the GNU Lesser General Public
    License along with the m17n library; if not, write to the Free
-   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    02111-1307, USA.  */
 
 #ifndef _M17N_GUI_H_
 #define _M17N_GUI_H_
 
+#ifndef _M17N_FLT_H_
+#include <m17n-flt.h>
+#endif
+
 #ifndef _M17N_H_
 #include <m17n.h>
 #endif
 
-#ifdef __cplusplus
-extern "C"
-{
-#endif
+M17N_BEGIN_HEADER
 
 #if !defined (FOR_DOXYGEN) || defined (DOXYGEN_INTERNAL_MODULE)
 
@@ -410,15 +411,11 @@ typedef struct
 /***en
     @brief Type of hook function of face.
 
-    The type #MFaceHookFunc is to specify the #Mhook property of a
-    face.  The value of the property must be function of this
-    type.  */
+    #MFaceHookFunc is a type of a hook function of a face.  */
 /***ja
     @brief եΥեåؿη.
 
-    #MFaceHookFunc ϥե #Mhook ץѥƥꤹ뷿Ǥ롣
-    ΥץѥƥͤϡηδؿǤʤƤϤʤʤ
-      */
+    #MFaceHookFunc ϥեΥեåؿηǤ롣*/
 typedef void (*MFaceHookFunc) (MFace *face, void *arg, void *info);
 /*=*/
 
@@ -426,6 +423,10 @@ extern void *mface_get_prop (MFace *face, MSymbol key);
 
 extern int mface_put_prop (MFace *face, MSymbol key, void *val);
 
+extern MFaceHookFunc mface_get_hook (MFace *face);
+
+extern int mface_put_hook (MFace *face, MFaceHookFunc func);
+
 extern void mface_update (MFrame *frame, MFace *face);
 
 /* end of face module */
@@ -987,9 +988,7 @@ extern MFace *mdebug_dump_face (MFace *face, int indent);
 extern MFont *mdebug_dump_font (MFont *font);
 extern MFontset *mdebug_dump_fontset (MFontset *fontset, int indent);
 
-#ifdef __cplusplus
-}
-#endif
+M17N_END_HEADER
 
 #endif /* _M17N_GUI_H_ */
 
