@@ -17,7 +17,7 @@
 
    You should have received a copy of the GNU Lesser General Public
    License along with the m17n library; if not, write to the Free
-   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    02111-1307, USA.  */
 
 /***en
@@ -155,7 +155,7 @@ msymbol__fini ()
       if (! MPLIST_TAIL_P (&sym->plist))
 	{
 	  if (sym->plist.key->managing_key)
-	    M17N_OBJECT_UNREF (sym->plist.val);
+	    M17N_OBJECT_UNREF (MPLIST_VAL (&sym->plist));
 	  M17N_OBJECT_UNREF (sym->plist.next);
 	  sym->plist.key = Mnil;
 	}
@@ -179,7 +179,7 @@ msymbol__free_table ()
 	}
       symbol_table[i] = NULL;
     }
-  if (mdebug__flag & MDEBUG_FINI)
+  if (mdebug__flags[MDEBUG_FINI])
     fprintf (stderr, "%16s %7d %7d %7d\n", "Symbol",
 	     num_symbols, freed_symbols, num_symbols - freed_symbols);
   num_symbols = 0;
@@ -660,6 +660,59 @@ msymbol_get (MSymbol symbol, MSymbol key)
   return (MPLIST_TAIL_P (plist) ? NULL : MPLIST_VAL (plist));
 }
 
+/*=*/
+/***en
+    @brief Set the value (function pointer) of a symbol property.
+
+    The msymbol_put_func () function is similar to msymbol_put () but for
+    setting function pointer $FUNC as the property value of $SYMBOL for
+    key $KEY.  */
+
+/***ja
+    @brief ܥץѥƥ(ؿݥ)ꤹ.
+
+    ؿ msymbol_put_func () ϡؿ msymbol_put () Ʊͤˡܥ
+    $SYMBOL Υ $KEY Ǥ륷ܥץѥƥͤꤹ롣â
+    ͤϴؿݥ $FUNC Ǥ롣 */
+
+/***
+    @seealso msymbol_put (), M17N_FUNC ()  */
+int
+msymbol_put_func (MSymbol symbol, MSymbol key, M17NFunc func)
+{
+  if (symbol == Mnil || key == Mnil)
+    MERROR (MERROR_SYMBOL, -1);
+  mplist_put_func (&symbol->plist, key, func);
+  return 0;
+}
+
+/*=*/
+
+/***en
+    @brief Get the value (function pointer) of a symbol property.
+
+    The msymbol_get_func () function is similar to msymbol_get () but for
+    getting a function pointer form the property of symbol $SYMBOL.  */
+
+/***ja
+    @brief ܥץѥƥ (ؿݥ) .
+
+    ؿ msymbol_get_func () ϡؿ msymbol_get () Ʊͤˡܥ
+    $SYMBOL ĥܥץѥƥΤ $KEY ǤΤ롣â
+    ͤϴؿݥ󥿤Ǥ롣    */
+
+/***
+    @seealso
+    msymbol_get ()  */
+
+M17NFunc
+msymbol_get_func (MSymbol symbol, MSymbol key)
+{
+  if (symbol == Mnil || key == Mnil)
+    return NULL;
+  return mplist_get_func (&symbol->plist, key);
+}
+
 /*** @} */
 
 #include <stdio.h>
